#!python
"""
	Main script to run the muller workflow.
"""
from pathlib import Path

from loguru import logger

logger.remove()
import sys
sys.path.append(str(Path(__file__).parent.parent)) # To deal with import errors.
logger.add(sys.stderr, level = "INFO")
# logger.add("muller_log.txt", level = 'DEBUG')
try:
	from muller import dataio, clustering, inheritance, commandline_parser
	from muller.muller_output import WorkflowData, generate_output
except ModuleNotFoundError:
	from . import dataio, clustering, inheritance, commandline_parser
	from .muller_output import WorkflowData, generate_output


def workflow(input_filename: Path, output_folder: Path, program_options):
	# as long as the sum of the other muller_genotypes that inherit from root is less than 1.
	logger.info("parsing options...")
	program_options = commandline_parser.parse_workflow_options(program_options)
	logger.info("Program options:")
	for k, v in vars(program_options).items():
		logger.info(f"\t{k:<20}{v}")

	logger.info("Importing data...")
	if program_options.known_ancestry:
		logger.info(f"Attempting to load ancestry data...")
		known_ancestry = dataio.read_map(program_options.known_ancestry)
	else:
		known_ancestry = dict()
	breakpoints = program_options.frequencies if program_options.use_filter else None
	genotype_generator = clustering.generate_genotypes.ClusterMutations(
		method = program_options.method,
		metric = program_options.metric,
		dlimit = program_options.detection_breakpoint,
		flimit = program_options.fixed_breakpoint,
		sbreakpoint = program_options.similarity_breakpoint,
		dbreakpoint = program_options.detection_breakpoint,
		breakpoints = breakpoints,
		starting_genotypes = program_options.starting_genotypes
	)
	if program_options.is_genotype:
		mean_genotypes, genotype_info = dataio.parse_genotype_table(input_filename, program_options.sheetname)
		try:
			genotype_members = genotype_info['members']
		except KeyError:
			genotype_members = dict()
		timepoints = info = None

	else:
		timepoints, info = dataio.parse_trajectory_table(input_filename, program_options.sheetname)
		mean_genotypes, genotype_members = genotype_generator.run(timepoints)


	logger.info("sorting muller_genotypes...")
	sorted_genotypes = inheritance.sort_genotypes(
		mean_genotypes,
		dlimit = program_options.detection_breakpoint,
		slimit = program_options.significant_breakpoint,
		flimit = program_options.fixed_breakpoint,
		breakpoints = program_options.frequencies
	)
	logger.info("nesting muller_genotypes...")

	genotype_clusters = inheritance.order_clusters(
		sorted_genotypes,
		dlimit = program_options.detection_breakpoint,
		flimit = program_options.fixed_breakpoint,
		additive_cutoff = program_options.additive_cutoff,
		subtractive_cutoff = program_options.subtractive_cutoff,
		derivative_cutoff = program_options.derivative_cutoff,
		known_ancestry = known_ancestry
	)
	logger.info("Generating output...")
	# TODO Make 'genotype-0' a variable rather than hard-coding it.
	# TODO Add some options to control how the graphics are generated. Ex. the outlines.
	workflow_data = WorkflowData(
		version = commandline_parser.__VERSION__,
		filename = input_filename,

		info = info,
		original_trajectories = timepoints,
		original_genotypes = mean_genotypes,
		rejected_trajectories = genotype_generator.rejected_trajectories,
		trajectories = timepoints,
		genotypes = sorted_genotypes,
		genotype_members = genotype_members,
		clusters = genotype_clusters,
		program_options = vars(program_options),
		p_values = genotype_generator.pairwise_distances,
		filter_cache = [],
		linkage_matrix = genotype_generator.linkage_table,
		genotype_palette_filename = program_options.genotype_palette_filename
	)
	generate_output(
		workflow_data,
		output_folder,
		program_options.detection_breakpoint,
		adjust_populations = True
	)
	return genotype_clusters


if __name__ == "__main__":
	args = commandline_parser.create_parser().parse_args()

	workflow(args.filename, args.output_folder, program_options = args)



from typing import List, Optional


class PopulationGenotypes:
	def __init__(self, genotypes: List[List[str]] = None):
		if genotypes is None:
			genotypes = []
		self.genotypes = dict()
		for genotype in genotypes:
			self.create_genotype(genotype)

	def __len__(self) -> int:
		largest_index = max(self.genotypes.keys(), key = lambda s: int(s.split('-')[-1]))
		return int(largest_index.split('-')[-1])

	def to_list(self) -> List[List[str]]:
		return list(v for k, v in sorted(self.genotypes.items()))

	def merge_trajectories(self, left: str, right: str):
		genotype_left = self.get_genotype_from_trajectory(left)
		genotype_right = self.get_genotype_from_trajectory(right)

		if genotype_left and genotype_right:
			self.merge_genotypes(genotype_left, genotype_right)
		elif genotype_left:
			self.add_to_genotype(genotype_left, right)
		elif genotype_right:
			self.add_to_genotype(genotype_right, left)
		else:
			self.create_genotype([left, right])

	def add_to_genotype(self, genotype_label: str, new_member: str):
		self.genotypes[genotype_label].append(new_member)

	def create_genotype(self, members: List[str]):
		genotype_index = len(self.genotypes) + 1
		genotype_label = f"genotype-{genotype_index}"
		self.genotypes[genotype_label] = members

	def merge_genotypes(self, genotype_a: str, genotype_b: str):
		if genotype_a == genotype_b: return None
		genotype_a_members = self.genotypes[genotype_a]
		genotype_b_members = self.genotypes[genotype_b]

		new_genotype = genotype_a_members + genotype_b_members
		from pprint import pprint
		pprint(self.genotypes)
		self.create_genotype(new_genotype)

		self.genotypes.pop(genotype_a)
		self.genotypes.pop(genotype_b)

	def get_genotype_from_trajectory(self, trajectory_label: str) -> Optional[str]:
		for genotype_label, genotype_members in self.genotypes.items():
			if trajectory_label in genotype_members:
				return genotype_label
		return None



from typing import List

import pandas
from loguru import logger

try:
	from muller.inheritance import timepoint_detection
except ModuleNotFoundError:
	from . import timepoint_detection


def sort_genotypes(genotype_frequencies: pandas.DataFrame, dlimit: float, slimit: float, flimit: float, breakpoints: List[float]) -> pandas.DataFrame:
	"""
		Sorts the muller_genotypes based on when they were first detected and first fixed.
	Parameters
	----------
	genotype_frequencies:pandas.Dataframe
		A dataframe with the mean frequency of each genotype, derived from the member trajectories
		in that genotype. Each row should correspond to a single genotype.
	dlimit, slimit, flimit: float
		The three breakpoints that are used to determine the sort order of the genotypes.
	breakpoints: List[float]
		Frequencies with which to group genotypes into sortable bins. Each bin will be sorted individually then added to the final output.

	Returns
	-------
	sorted_genotype: pandas.DataFrame
	"""
	sorted_genotypes = list()
	current_genotypes: pandas.DataFrame = genotype_frequencies.copy()
	for frequency in [flimit] + breakpoints:
		logger.debug(f"filtering based on frequency {frequency}")
		# Ignore genotypes that do not have at least on timepoint exceeding the current frequency.
		genotypes_above_threshold = _remove_low_frequency_series(current_genotypes, frequency)

		sorted_dataframe = _sort_genotype_frequencies(
			genotype_trajectories = genotypes_above_threshold,
			frequency_breakpoint = frequency,
			detection_cutoff = dlimit,
			significant_cutoff = slimit,
			fixed_cutoff = flimit
		)
		if sorted_dataframe is not None:
			current_genotypes = current_genotypes.drop(sorted_dataframe.index)
			sorted_genotypes.append(sorted_dataframe)
	df = pandas.concat(sorted_genotypes, sort = False)
	df = genotype_frequencies.reindex(df.index)
	return df


def _remove_low_frequency_series(df: pandas.DataFrame, threshold: float) -> pandas.DataFrame:
	""" Removes all genotypes that do not have at least one frequency above the given threshold."""
	result = df[df.max(axis = 1) >= threshold]
	return result


# noinspection PyUnresolvedReferences,PyTypeChecker
def _get_timepoint_above_threshold(transposed_timepoints: pandas.DataFrame, cutoff: float, name: str = None) -> pandas.Series:
	"""
		Calculates when a genotype was first fixed based on `cutoff`. The resulting series will be named `name`, if given.
	Parameters
	----------
	transposed_timepoints: pandas.DataFrame
		A dataframe indexed by timepoint and with each series a separate column.
	cutoff: float
		The threshold a frequency must exceed to be considered a valid timepoint.

	Returns
	-------
	pandas.Series
		A series mapping series name to the first timepoint that series first exceeded the threshold value.
		Series that never exceed the threshold are mapped to a value of 0.
	"""
	threshold_series: pandas.Series = (transposed_timepoints > cutoff).idxmax(0).sort_values()
	if name:
		threshold_series.name = name
	return threshold_series


def _sort_genotype_frequencies(genotype_trajectories: pandas.DataFrame, frequency_breakpoint: float,
		detection_cutoff: float, significant_cutoff: float, fixed_cutoff: float) -> pandas.DataFrame:
	"""
		Sorts the genotype table based on the timepoints that first satisfy the threshold requirements.
	Parameters
	----------
	genotype_trajectories
	frequency_breakpoint
	detection_cutoff
	significant_cutoff
	fixed_cutoff

	Returns
	-------
	"""
	# Get the values for each genotype at the first timepoint. This will be used to check if the genotype existed at the first timepoint
	# or has an initial detected timepoint of 0 due to the way pandas returns indicies.
	detection_cutoff = min(detection_cutoff, frequency_breakpoint)
	transposed = genotype_trajectories.transpose()

	first_detected_reduced = timepoint_detection.get_first_detected_timepoint(transposed, detection_cutoff)
	first_above_threshold_reduced = timepoint_detection.get_first_significant_timepoint(transposed, significant_cutoff)
	# Use the frequency breakpoint rather than the fixed cutoff.
	first_fixed_reduced = timepoint_detection.get_first_fixed_timepoint(transposed, frequency_breakpoint)

	combined_df: pandas.DataFrame = pandas.concat(
		[first_fixed_reduced, first_detected_reduced, first_above_threshold_reduced],
		axis = 1, sort = False)
	df = combined_df[~combined_df['firstFixed'].isna()]

	if frequency_breakpoint == fixed_cutoff:
		df = df.dropna()
	sorted_frequencies = df.sort_values(by = ['firstDetected', 'firstThreshold'], ascending = False)

	# Sort genotypes based on frequency if two or more share the same fixed timepoint, detection timpoint, threshold timepoint.
	# Iterate over the conbinations of 'firstFixed', 'firstDetected', and 'firstThreshold' and sort trajectories that belong to the sample combination.
	# Need to build a new trajectory table based on the sorted timepoint table.
	freq_df = _build_sorted_frequency_table(genotype_trajectories, sorted_frequencies)

	return freq_df


def _build_sorted_frequency_table(original_frequencies: pandas.DataFrame, thresholds: pandas.DataFrame) -> pandas.DataFrame:
	# Sort genotypes based on frequency if two or more share the same fixed timepoint, detection timpoint, threshold timepoint.
	# Iterate over the conbinations of 'firstFixed', 'firstDetected', and 'firstThreshold' and sort trajectories that belong to the sample combination.
	# Need to build a new trajectory table based on the sorted timepoint table.

	freq_groups = list()
	# Group the
	groups = thresholds.groupby(by = list(thresholds.columns))
	for (ff, fd, ft), group in groups:
		# Each group will look like this:
		#   	firstFixed firstDetected firstThreshold
		# 15          0            25              0
		# 10          0            25              0
		if ft == 130:  # dummy value assigned above. Revert back to 0 since '130' isn't a valid timepoint.
			ft = 0
		# Get a table of the original frequencies at each timepoint for the genotypes present in `group`
		trajectories: pandas.DataFrame = original_frequencies.loc[group.index]
		if len(group) < 2:
			# There is only one genotype in this combination of timepoints. No need to sort.
			freq_groups.append(trajectories)
		else:
			# More than one genotype share this combination of key timepoints. Sort by frequency.
			# Sort from highest to lowest using the timpoint columns as the sorting keys.
			# trajectories = trajectories.sort_values(by = [ff, ft, fd], ascending = False)
			trajectories = trajectories.sort_values(by = [fd, ft, ff], ascending = False)
			freq_groups.append(trajectories)

	try:
		freq_df = pandas.concat(freq_groups)
	except ValueError:
		# Can't concatenate an empty dataframe
		freq_df = None

	return freq_df



import io
from pathlib import Path
from typing import Optional, Union

import pandas


def _import_table_from_path(filename: Path, sheet_name: Optional[str] = None, index: Optional[str] = None) -> pandas.DataFrame:
	""" Imports a file as a pandas.DataFrame. Infers filetype from the filename extension/suffix.
	"""
	if filename.suffix in {'.xls', '.xlsx'}:
		data: pandas.DataFrame = pandas.read_excel(str(filename), sheet_name = sheet_name)
	else:
		sep = '\t' if filename.suffix in {'.tsv', '.tab'} else ','
		data: pandas.DataFrame = pandas.read_csv(str(filename), sep = sep)

	if index and index in data.columns:
		data = data.set_index(index)

	return data


def _import_table_from_string(string: str, delimiter: Optional[str] = None, index: Optional[str] = None) -> pandas.DataFrame:
	""" Imports a table represented as a basic string object."""
	# Remove unwanted whitespace.
	string = '\n'.join(i.strip() for i in string.split('\n') if i)
	if not delimiter:
		delimiter = '\t' if '\t' in string else ','
	result = pandas.read_csv(io.StringIO(string), sep = delimiter, index_col = False)
	if index:
		# Using `index_col` in `read_table()` doesn't work for some reason.
		result[index] = result[index].astype(str)
		result.set_index(index, inplace = True)
	return result


def import_table(input_table: Union[str, Path], sheet_name: Optional[str] = None, index: Optional[str] = None) -> pandas.DataFrame:
	if isinstance(input_table, Path):
		data = _import_table_from_path(input_table, sheet_name, index)
	else:
		data = _import_table_from_string(input_table, index = index)
	return data



import itertools
import math
from typing import Dict, Iterable, Tuple

import pandas
from loguru import logger

try:
	from muller.clustering.metrics import distance
	from muller import widgets
except ModuleNotFoundError:
	from . import distance
	from ... import widgets
try:
	from tqdm import tqdm
except ModuleNotFoundError:
	tqdm = None

def fixed_overlap(left: pandas.Series, right: pandas.Series, fixed_cutoff: float) -> float:
	"""
		Calculates the overlap of two series that are both undetected or fixed at all timepoints.
	Parameters
	----------
	left:pandas.Series
	right:pandas.Series
	fixed_cutoff: float

	Returns
	-------
	float
	"""
	left_fixed: pandas.Series = left[left.gt(fixed_cutoff)].dropna()
	right_fixed: pandas.Series = right[right.gt(fixed_cutoff)].dropna()
	if left_fixed.empty or right_fixed.empty:
		# Both are undetected, since they both failed the invalid timepoint filter.
		p_value = False
	else:
		# Since these genotypes fixed immediately, they should only be grouped together if they
		# fixed at the same timepoint.
		fixed_at_same_time = left_fixed.index[0] == right_fixed.index[0]
		# Check if the trajectories overlap completely
		overlap = set(left_fixed.index) & set(right_fixed.index)
		complete_overlap = len(overlap) == len(left_fixed) and len(overlap) == len(right_fixed)
		p_value = fixed_at_same_time and complete_overlap
	result = 0 if p_value else math.nan

	return result


def calculate_pairwise_metric(trajectories: pandas.DataFrame, detection_cutoff: float, fixed_cutoff: float, metric: str) -> Dict[
	Tuple[str, str], float]:
	"""
	Parameters
	----------
	trajectories: pandas.DataFrame
		A table of mutational trajectories. Should be a normal trajectory table.
	detection_cutoff: float
	fixed_cutoff: float
	metric: {'similarity', 'binomial', 'pearson', 'minkowski'}

	Returns
	-------
	dict of PairArrayValue
	Each key in the dictionary corresponds to a pair of trajectory ids which map to the p-value for that pair.
	The order of ids does not matter.
	"""
	logger.debug("Calculating the pairwise values...")
	logger.debug(f"\t detection limit: {detection_cutoff}")
	logger.debug(f"\t fixed limit: {fixed_cutoff}")
	logger.debug(f"\t metric: {metric}")

	# noinspection PyTypeChecker
	pair_combinations: Iterable[Tuple[str, str]] = list(itertools.combinations(trajectories.index, 2))
	pair_array = dict()
	if len(pair_combinations) > 10000 and tqdm:
		progress_bar = tqdm(total = len(pair_combinations))
	else:
		progress_bar = None
	for left, right in pair_combinations:
		if progress_bar:
			progress_bar.update(1)
		left_trajectory = trajectories.loc[left]
		right_trajectory = trajectories.loc[right]

		# We only care about the timepoints such that `detection_cutoff` < f < `fixed_cutoff.
		# For now, lets require that both timepoints are detected and not yet fixed.
		# There is an issue related to comparing fixed genotypes against non-fixed genotypes.
		# These were grouped together:
		# recG    0    0.14    0.319    1    1    1    1
		# PA14_RS20565< 0    0.153    0.231    0    0    0    0
		#left_reduced, right_reduced = widgets.get_valid_points(left_trajectory, right_trajectory, detection_cutoff, fixed_cutoff, inner = False)
		left_was_fixed = widgets.fixed(left_trajectory, fixed_cutoff)
		right_was_fixed = widgets.fixed(right_trajectory, fixed_cutoff)
		if left_was_fixed == right_was_fixed:
			left_reduced, right_reduced = widgets.get_valid_points(left_trajectory, right_trajectory, detection_cutoff, fixed_cutoff, inner = False)
		else:
			left_reduced, right_reduced = widgets.get_valid_points(left_trajectory, right_trajectory, detection_cutoff, inner = False)

		if left_reduced.empty or right_reduced.empty:
			# Treat both trajectories as fixed immediately.
			distance_between_series = fixed_overlap(left_trajectory, right_trajectory, fixed_cutoff)
		else:
			distance_between_series = distance.calculate_distance(left_reduced, right_reduced, metric)

		pair_array[left, right] = pair_array[right, left] = distance_between_series
	if progress_bar: progress_bar.close()
	# Assume that any pair with NAN values are the maximum possible distance from each other.

	maximum_distance = max(filter(lambda s: not math.isnan(s), pair_array.values()))
	pair_array = {k: (v if not math.isnan(v) else maximum_distance) for k, v in pair_array.items()}

	return pair_array



from collections import OrderedDict
from typing import Dict, List

from . import colorset


def generate_distinctive_palette(genotypes: List[str]) -> Dict[str, str]:
	""" Assigns a unique color to each genotype."""
	color_palette = colorset.get_distinctive_palette(len(genotypes))

	genotype_labels = sorted(genotypes, key = lambda s: int(s.split('-')[-1]))
	# Use an OrderedDict to help with providing the correct order for the r script.
	color_map = OrderedDict()
	for label, color in zip(genotype_labels, color_palette):
		color_map[label] = color

	return color_map



import itertools
import math
from typing import Dict, Iterable, List, Tuple

import pandas

try:
	from muller.clustering.metrics import distance
	from muller.widgets import get_valid_points
except ModuleNotFoundError:
	from . import distance
	from ..widgets import get_valid_points


class CalculateMutationalDistances:
	def __init__(self, metric: str, detection_cutoff: float, fixed_cutoff: float):
		self.dlimit = detection_cutoff
		self.flimit = fixed_cutoff
		self.metric = metric

	def run(self, trajectories: pandas.DataFrame) -> Dict[Tuple[str, str], float]:
		pairwise_keys = self._generate_pairwise_combinations(trajectories.index)
		pair_array = dict()
		for left, right in pairwise_keys:
			left_trajectory = trajectories.loc[left]
			right_trajectory = trajectories.loc[right]

			# We only care about the timepoints such that `detection_cutoff` < f < `fixed_cutoff.
			# For now, lets require that both timepoints are detected and not yet fixed.
			left_reduced, right_reduced = get_valid_points(left_trajectory, right_trajectory, self.dlimit, self.flimit, inner = False)

			if left_reduced.empty or right_reduced.empty:
				distance_between_series = self.calculate_fixed_series_distance(left_trajectory, right_trajectory)
			else:
				distance_between_series = distance.calculate_distance(left_reduced, right_reduced, self.metric)

			pair_array[left, right] = pair_array[right, left] = distance_between_series

		return pair_array

	@staticmethod
	def _generate_pairwise_combinations(keys: Iterable[str]) -> List[Tuple[str, str]]:
		""" Generates a list of all possible pair of elements contained in `keys`"""

		pair_combinations: Iterable[Tuple[str, str]] = itertools.combinations(keys, 2)
		return list(pair_combinations)

	@staticmethod
	def _get_maximum_value(items: Iterable[float]) -> float:
		""" Extracts the maximum values contained within an iterable."""
		return max(filter(lambda s: not math.isnan(s), items))

	def calculate_fixed_series_distance(self, left: pandas.Series, right: pandas.Series) -> float:
		"""
			Calculates the overlap of two series that are both undetected or fixed at all timepoints.
		Parameters
		----------
		left:pandas.Series
		right:pandas.Series
		fixed_cutoff: float

		Returns
		-------
		float
		"""
		left_fixed: pandas.Series = left[left.gt(self.flimit)].dropna()
		right_fixed: pandas.Series = right[right.gt(self.flimit)].dropna()
		if left_fixed.empty or right_fixed.empty:
			# Both are undetected, since they both failed the invalid timepoint filter.
			p_value = False
		else:
			# Since these genotypes fixed immediately, they should only be grouped together if they
			# fixed at the same timepoint.
			fixed_at_same_time = left_fixed.index[0] == right_fixed.index[0]
			# Check if the trajectories overlap completely
			overlap = set(left_fixed.index) & set(right_fixed.index)
			complete_overlap = len(overlap) == len(left_fixed) and len(overlap) == len(right_fixed)
			p_value = fixed_at_same_time and complete_overlap
		result = 0 if p_value else math.nan

		return result



from setuptools import setup
from muller import commandline_parser
from pathlib import Path

FOLDER = Path(__file__).parent
README = FOLDER / "README.md"

with README.open() as readmefile:
    LONG_DESCRIPTION = readmefile.read()

setup(
	name = 'muller',
	version = commandline_parser.__VERSION__,
	packages = [
		'muller', 'muller.clustering', 'muller.dataio', 'muller.graphics', 'muller.inheritance',
		'muller.muller_output', 'muller.palettes', 'muller.clustering.metrics', 'muller.clustering.methods'
	],
	provides = 'muller',
	url = 'https://github.com/cdeitrick/muller_diagrams',
	license = 'MIT',
	author = 'chris deitrick',
	author_email = 'chrisdeitrick1@gmail.com',
	description = 'A set of scripts to cluster mutational trajectories into genotypes and cluster genotypes by background',
	long_description = LONG_DESCRIPTION,
	long_description_content_type='text/markdown',
	install_requires = ['pandas', 'loguru', 'scipy', 'matplotlib','graphviz', 'pygraphviz', 'seaborn', 'numpy', 'xlrd'],
	tests_requires = ['pytest'],
	classifiers = [
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
	scripts = ["Muller.py"]
)



from typing import List, Optional, Tuple

import numpy
import pandas
from loguru import logger
try:
	from muller.clustering import filters, metrics, methods
except ModuleNotFoundError:
	from . import filters, metrics, methods


class ClusterMutations:
	"""
	Parameters
	----------
	trajectories: pandas.DataFrame
		A timeseries dataframe, usually generated from `import_table.import_trajectory_table`.
			- Index -> str
				Names unique to each trajectory.
			- Columns -> int
				The timeseries points will correspond to the frequencies for each trajectory included with the input sheet.
				Each trajectory/timepoint will include the observed frequency at each timepoint.
	dlimit, flimit:float
		The detection, significant, and fixed cutoff values, respectively
	sbreakpoint: float
		The cutoff value to use when clustering trajectories into genotypes. Two trajectories must have a distance less than this value to be
		considered members of the same genotype.
	dbreakpoint: float
		Only used when using the two-step method. Governs when the genotype contains two trajectories which should be split into separate genotypes.
	method: {'two-step', 'hierarchy'}
		The clustering method to use. The hierarchical method uses the maximum distance between a point and a candidate cluster to determine whether
		to group said point to the cluster.
	metric: {'binomial', 'pearson', 'minkowski'}
		The distance metric to determine trajectory similarity.
	starting_genotypes: List[List[str]]
		A list of genotypes to start the clustering algorithm with. The distance metrics will be modified so that the trajectories specified are
		grouped together.
	Returns
	-------
	pandas.DataFrame, pandas.Series, numpy.array
		- The genotype table
		- A map of genotypes to members.
		- A linkage matrix, if hierarchical clustering was used.
	"""

	def __init__(self, method: str, metric: str, dlimit: float, flimit: float, sbreakpoint: float, dbreakpoint: float, breakpoints: List[float],
			starting_genotypes: List[List[str]]):
		self.method: str = method
		self.metric: str = metric
		self.dlimit: float = dlimit
		self.flimit: float = flimit
		self.sbreakpoint: float = sbreakpoint
		self.dbreakpoint: float = dbreakpoint
		self.breakpoints: List[float] = breakpoints
		self.starting_genotypes: List[List[str]] = starting_genotypes

		# These will be updated when run() is called.
		self.pairwise_distances: metrics.PairwiseCalculationCache = metrics.PairwiseCalculationCache()  # Empty cache that will be replaced in generate_genotypes().
		self.genotype_table = self.genotype_members = self.linkage_table = self.rejected_trajectories = None

	def run(self, trajectories: pandas.DataFrame):
		trajectory_filter = filters.TrajectoryFilter(detection_cutoff = self.dlimit, fixed_cutoff = self.flimit)
		genotype_filter = filters.GenotypeFilter(detection_cutoff = self.dlimit, fixed_cutoff = self.flimit, frequencies = self.breakpoints)
		modified_trajectories = trajectories.copy(deep = True)  # To avoid unintended changes
		if self.breakpoints:
			# The filters should run
			_iterations = 20  # arbitrary, used to make sure the program does not encounter an infinite loop.
			modified_trajectories = trajectory_filter.run(modified_trajectories)
		else:
			# The filters are disabled.
			_iterations = 0  # The for loop shouldn't excecute.

		# Calculate the distance between all possible pair of trajectories.
		pair_array = metrics.calculate_pairwise_metric(
			modified_trajectories,
			detection_cutoff = self.dlimit,
			fixed_cutoff = self.flimit,
			metric = self.metric
		)
		self.pairwise_distances = metrics.PairwiseCalculationCache(pair_array)

		# Calculate the initial genotypes
		genotype_table, genotype_members, linkage_matrix = self.generate_genotypes(modified_trajectories)

		rejected_members = {i:'filtered-genotype-0' for i in trajectories.index if i not in modified_trajectories}
		for index in range(_iterations):
			invalid_members = genotype_filter.run(genotype_table, genotype_members)
			if invalid_members:
				for i in invalid_members:
					rejected_members[i] = f"filtered-genotype-{index+1}"
				# Remove these trajectories from the trajectories table.
				modified_trajectories = modified_trajectories[~modified_trajectories.index.isin(invalid_members)]

				# Need to remove trajectories from the distance matrix so they are not included in the clustering method.
				self.pairwise_distances.reduce(modified_trajectories.index)
				# Re-calculate the genotypes based on the remaining trajectories.

				genotype_table, genotype_members, linkage_matrix = self.generate_genotypes(modified_trajectories)
			else:
				break
		# Build a table of trajectories that were rejected.
		self.rejected_trajectories = trajectories.loc[sorted(rejected_members.keys())]
		self.rejected_trajectories['genotype'] = [rejected_members[i] for i in self.rejected_trajectories.index]

		self.genotype_table = genotype_table
		self.genotype_members = genotype_members
		self.linkage_table = linkage_matrix

		return genotype_table, genotype_members

	def generate_genotypes(self, timepoints: pandas.DataFrame) -> Tuple[pandas.DataFrame, pandas.Series, Optional[numpy.array]]:
		if self.method == "matlab" or self.method == 'twostep':
			genotypes = methods.twostep_method(timepoints, self.pairwise_distances, self.sbreakpoint, self.dbreakpoint, self.starting_genotypes)
			linkage_matrix = None
		elif self.method == "hierarchy":
			genotypes, linkage_matrix = methods.hierarchical_method(self.pairwise_distances, self.sbreakpoint, starting_genotypes = self.starting_genotypes)
		else:
			raise ValueError(f"Invalid clustering method: {self.method}")

		mean_genotypes = self.calculate_mean_genotype(genotypes, timepoints)
		genotype_members = mean_genotypes.pop('members')
		return mean_genotypes, genotype_members, linkage_matrix

	@staticmethod
	def _calculate_mean_frequencies_of_trajectories(name: str, genotype_timeseries: pandas.DataFrame, genotype: List[str]) -> pandas.Series:
		"""
			Generates a mean timeseries for a genotype given the timeseries of member trajectories.
		Parameters
		----------
		name: str
			The name to assign to this genotype. Usually just 'genotype-' forllowed by an integer.
		genotype_timeseries: pandas.DataFrame
			A dataframe of member trajectory timeseries.
		genotype: List[str]
			A list of the labels of all member trajectories. Used to track the which trajectories belong to this genotype.

		Returns
		-------
		pandas.DataFrame
			The calculated mean of all member trajectories at each timepoint.
		"""
		mean_genotype_timeseries = genotype_timeseries.mean()
		mean_genotype_timeseries['members'] = "|".join(map(str, genotype))
		mean_genotype_timeseries.name = name

		return mean_genotype_timeseries

	def calculate_mean_genotype(self, all_genotypes: List[List[str]], timeseries: pandas.DataFrame) -> pandas.DataFrame:
		"""
			Calculates the mean frequency of each genotype ate every timepoint.
		Parameters
		----------
		all_genotypes: List[List[str]
			A list of all muller_genotypes for a given population
		timeseries: pandas.DataFrame
			Must have a 'Trajectory' column along with the columns of the original table that represent timepoints.
			Each row corresponds to a single mutational trajectory.

		Returns
		-------
		A dataframe where every row corresponds to a genotype.
		member trajectories are listed under the 'members' column.
		every column represents a timepoint.
		"""
		mean_genotypes = list()
		for index, genotype in enumerate(all_genotypes, start = 1):
			try:
				genotype_timeseries = timeseries.loc[genotype]
			except Exception as exception:
				logger.critical(f"Missing Trajectory Labels: {genotype} - {timeseries.index}")
				raise exception
			mean_genotype_timeseries = self._calculate_mean_frequencies_of_trajectories(f"genotype-{index}", genotype_timeseries, genotype)
			mean_genotypes.append(mean_genotype_timeseries)

		mean_genotypes = pandas.DataFrame(mean_genotypes)
		# For consistency
		mean_genotypes.index.name = 'Genotype'

		return mean_genotypes



"""
	Python implementation of the Muller_plot function available from ggmuller.
"""

import math
import random
from itertools import filterfalse
from pathlib import Path
from typing import Any, Dict, List, Tuple, Union

import pandas
from matplotlib import pyplot as plt
# plt.switch_backend('agg')
from matplotlib.figure import Axes  # For autocomplete

# Make sure the svg files save the labels as actualt text.
plt.rcParams['svg.fonttype'] = 'none'
try:
	from muller.widgets import calculate_luminance
except ModuleNotFoundError:
	from ..widgets import calculate_luminance

from loguru import logger

plt.style.use('seaborn-white')

pandas.set_option('display.max_rows', 500)
pandas.set_option('display.max_columns', 500)
pandas.set_option('display.width', 250)


def unique_everseen(iterable, key = None):
	"""	List unique elements, preserving order. Remember all elements ever seen."
		unique_everseen('AAAABBBCCDAABBB') --> A B C D
		unique_everseen('ABBCcAD', str.lower) --> A B C D
	"""
	seen = set()
	seen_add = seen.add
	if key is None:
		for element in filterfalse(seen.__contains__, iterable):
			seen_add(element)
			yield element
	else:
		for element in iterable:
			k = key(element)
			if k not in seen:
				seen_add(k)
				yield element


def generate_muller_series(muller_df: pandas.DataFrame, color_palette: Dict[str, str]) -> Tuple[List[float], List[List[float]], List[str], List[str]]:
	"""
		Generates the required inputs for matplotlib to generate a mullerplot.
	Parameters
	----------
	muller_df: pandas.DataFrame
	color_palette: Dict[str,str]
		Maps genotypes to a specific color.

	Returns
	-------
	x, y, colors, labels
	"""
	genotype_order = list(unique_everseen(muller_df['Group_id'].tolist()))

	x = list(unique_everseen(muller_df['Generation'].tolist()))
	labels = [(label if not label.endswith('a') else None) for label in genotype_order]

	groups = muller_df.groupby(by = 'Group_id')
	ys = list()
	colors = list()
	seen = set()
	for index, label in enumerate(genotype_order):
		if label in seen: continue
		seen.add(label)
		genotype_color = color_palette[label[:-1] if label.endswith('a') else label]
		try:
			next_label = genotype_order[index + 1]
		except IndexError:
			next_label = "   "  # Three spaces, so the following subscription works
		if label == next_label[:-1] and next_label[-1].isalpha():  # make sure it ends with a letter.
			# The two halves of this specific genotype are next to each other. combine them and skip the next iteration.
			seen.add(next_label)
			genotype_series_a = groups.get_group(label)['Frequency'].tolist()
			genotype_series_b = groups.get_group(next_label)['Frequency'].tolist()
			genotype_series = [i + j for i, j in zip(genotype_series_a, genotype_series_b)]
		else:
			# The genotype contains child genotypes.
			genotype_series = groups.get_group(label)['Frequency'].tolist()
		ys.append(genotype_series)
		colors.append(genotype_color)

	return x, ys, colors, labels


def get_coordinates(muller_df: pandas.DataFrame) -> Dict[str, Tuple[int, float]]:
	"""
		Attempts to find the mean point of the stacked area plot for each genotype.
	Parameters
	----------
	muller_df:pandas.DataFrame

	Returns
	-------
	points: Dict[str, Tuple[int, float]]
	A dictionary mapping each genotype to an estimate of the midpoint of the genotype's area.
	"""
	# TODO increase the font sizes of the labels
	genotype_order = list()
	for index, row in muller_df.iterrows():
		genotype_label = row['Group_id']
		if genotype_label not in genotype_order:
			genotype_order.append(genotype_label)

	nonzero = muller_df[muller_df['Population'] != 0]

	points = dict()
	groups = nonzero.groupby(by = 'Group_id')

	for index, name, in enumerate(genotype_order):
		genotype_label = name[:-1] if name.endswith('a') else name
		if genotype_label in points: continue
		try:
			group = groups.get_group(name)
		except KeyError:
			continue

		min_x = min(group['Generation'].values)
		max_x = max(group['Generation'].values)
		x = (min_x + max_x) / 2
		if x not in group['Generation'].values:
			x = sorted(group['Generation'].values, key = lambda s: s - x)[0]

		gdf = nonzero[nonzero['Group_id'].isin(genotype_order[:index] + [genotype_label])]
		gdf = gdf[gdf['Generation'] == x]

		mid_y = sum(gdf['Frequency'].values)
		if mid_y < 0: mid_y = 0
		if x == 0:
			x = 0.1
		points[genotype_label] = (x, mid_y)  # + .25)

	return points


def get_font_properties(genotype_color: str) -> Dict[str, Any]:
	"""
		Generates font properties for each annotations. The main difference is font color,
		which is determined by the background color.
	Parameters
	----------
	genotype_color: str
		The color of the genotype this text is associated with.

	Returns
	-------
	fontproperties: Dict[str, Any]
	"""

	luminance = calculate_luminance(genotype_color)
	if luminance > 0.5:
		font_color = "#333333"
	else:
		font_color = "#FFFFFF"
	label_properties = {
		'size':  9,
		'color': font_color
	}
	return label_properties


def distance(left: Tuple[float, float], right: Tuple[float, float]) -> float:
	xl, yl = left
	xr, yr = right

	num = (xl - xr) ** 2 + (yl - yr) ** 2
	return math.sqrt(num)


def find_closest_point(point, points: List[Tuple[float, float]]) -> Tuple[float, float]:
	return min(points, key = lambda s: distance(s, point))


def relocate_point(point: Tuple[float, float], locations: List[Tuple[float, float]]) -> Tuple[float, float]:
	x_loc, y_loc = point
	for _ in range(10):
		closest_neighbor = find_closest_point((x_loc, y_loc), locations)
		y_is_close = math.isclose(y_loc, closest_neighbor[1], abs_tol = .1)
		y_is_almost_close = math.isclose(y_loc, closest_neighbor[1], abs_tol = .3)
		x_is_close = math.isclose(x_loc, closest_neighbor[0], abs_tol = 2)
		x_large = x_loc >= closest_neighbor[0]
		logger.debug(f"{x_loc}\t{y_loc}\t{closest_neighbor}\t{y_is_close}\t{y_is_almost_close}\t{x_is_close}\t{x_large}")
		if distance(closest_neighbor, (x_loc, y_loc)) > 1: break
		if y_is_close:
			if y_loc > closest_neighbor[1]:
				y_loc += random.uniform(.1, .2)
			else:
				y_loc -= random.uniform(.05, .10)
		elif y_is_almost_close and x_is_close and x_large:
			x_loc += .5
			break
		else:
			break
	return x_loc, y_loc


def annotate_axes(ax: Axes, points: Dict[str, Tuple[float, float]], annotations: Dict[str, List[str]], color_palette: Dict[str, str]) -> Axes:
	locations = list()
	for genotype_label, point in points.items():
		background_properties = {
			'facecolor': color_palette[genotype_label],
			'alpha':     1,
			'edgecolor': "#FFFFFF",
			'linewidth': 0.5
		}
		if genotype_label == 'genotype-0': continue
		label_properties = get_font_properties(color_palette[genotype_label])

		if locations:
			x_loc, y_loc = relocate_point(point, locations)
		else:
			x_loc, y_loc = point
		locations.append((x_loc, y_loc))
		genotype_annotations = annotations.get(genotype_label, [])
		try:
			genotype_annotations = genotype_annotations
		except IndexError:
			pass
		ax.text(
			x_loc, y_loc,
			"\n".join(genotype_annotations),
			bbox = background_properties,
			fontdict = label_properties
		)
	return ax


def generate_muller_plot(muller_df: pandas.DataFrame, color_palette: Dict[str, str],
		output_filename: Union[Path, List[Path]], annotations: Dict[str, List[str]] = None, outlines:bool = True):
	"""
		Generates a muller diagram equivilent to r's ggmuller. The primary advantage is easier annotations.
	Parameters
	----------
	muller_df: pandas.DataFrame
	color_palette: Dict[str,str]
	output_filename: Path or List[Path]
		If a list of paths is given the same graphics will be saves as each filename. This is useful if you want to save the
		plot in a number of different formats.
	annotations: Dict[str, List[str]]
		A map of genotype labels to add to the plot.
	outlines: bool; default True
		Whether to add a white outline to the series in the muller plot.
	Returns
	-------
	ax: Axes
		The axes object containing the plot.
	"""
	# TODO add confidence to muller diagram and flowcharts.
	if not isinstance(output_filename, list):
		output_filenames = [output_filename]
	else:
		output_filenames = output_filename
	if outlines:
		edgecolor = '#FFFFFF'
	else:
		edgecolor = None

	points = get_coordinates(muller_df)

	# noinspection PyUnusedLocal,PyUnusedLocal
	fig, ax = plt.subplots(figsize = (12, 10))
	ax: Axes

	x, y, colors, labels = generate_muller_series(muller_df, color_palette)

	ax.stackplot(x, y, colors = colors, labels = labels, edgecolor = edgecolor, linewidth = 2, interpolate = True, joinstyle = 'round')
	if annotations:
		annotate_axes(ax, points, annotations, color_palette)
	else:
		plt.legend(loc = 'right', bbox_to_anchor = (1.05, 0.5, 0.1, 0), title = 'Identity')
	ax.set_facecolor("#FFFFFF")
	ax.set_xlabel("Generation", fontsize = 20)
	ax.set_ylabel("Frequency", fontsize = 20)

	# Basic stacked area chart.
	ax.spines['right'].set_visible(False)
	ax.spines['top'].set_visible(False)
	ax.set_xlim(0, max(x))
	# ax.set_ylim(0, 1)
	#plt.tight_layout()

	for output_filename in output_filenames:
		plt.savefig(str(output_filename))

	return ax
